/*
Code is partly based on pseudo code out of the paper:
"There Is No 16-Clue Sudoku: Solving the Sudoku Minimum Number of Clues Problem via Hitting Set Enumeration".
It is written by Gary McGuire, Bastian Tugemann, and Gilles Civario in 2014 and
published in Experimental Mathematics, 23:190217
*/

#include "SudokuGrid.h"
#include <cmath>
using namespace std;

unsigned int factorial(unsigned int n) {
 	unsigned int retval = 1;
 	for (int i = n; i > 1; --i)
 		retval *= i;
 	return retval;
}


// builds domain with initial values assigned
SudokuGrid::SudokuGrid(int size) : n(size) {
	n = size;
	grid.resize(pow(n,2));
	for (int var = 0; var < pow(n,2); ++var) {
		grid[var].resize(pow(n,2), 0);
	}
}

SudokuGrid::SudokuGrid(const SudokuGrid& copy, const vector<byte>& cells) : SudokuGrid(copy.n) {
	for (int i = 0; i < cells.size(); i++) {
		grid[cells[i] / pow(copy.n, 2)][cells[i] % (int)pow(copy.n, 2)] = copy.grid[cells[i] / pow(copy.n, 2)][cells[i] % (int)pow(copy.n, 2)];
	}
}

SudokuGrid::SudokuGrid(int size, const vector<byte>& cells, bool transformed) : SudokuGrid(size) {
	if (!transformed) {
		for (unsigned int i = 0; i < cells.size(); ++i) {
			grid[cells[i] / 10 - 1][cells[i] % 10 - 1] = 1;
		}
	}
	else {
		for (unsigned int i = 0; i < cells.size(); ++i) {
			grid[cells[i] / (int)pow(size,2)][cells[i] % (int)pow(size, 2)] = 1;
		}
	}
}

SudokuGrid::SudokuGrid(std::vector<byte> digits) : SudokuGrid(sqrt(sqrt(digits.size()))) {
	for (unsigned int i = 0; i < digits.size(); ++i) {
		if (digits[i] != 0) {
			grid[i/9][i%9] = digits[i];
		}
	}
}

SudokuGrid::SudokuGrid(int size, UnavPat pattern) : SudokuGrid(size) {
	for (int i = 1; i <= pattern.digits; i++) {
		for (int j = 0; j < pattern.groups[i-1].size; j++) {
			int row = pattern.groups[i - 1].data[j] / 10 - 1;
			int col = pattern.groups[i - 1].data[j] % 10 - 1;
			grid[row][col] = i;
		}
	}
}

SudokuGrid::SudokuGrid(const SudokuGrid& copy) /*: SudokuGrid(copy.n)*/  {
	/*for (int row = 0; row < pow(n,2); ++row) {
		for (int col = 0; col < pow(n,2); ++col) {
			grid[row][col] = copy[row][col];
		}
	}*/
	n = copy.n;
	grid.reserve(copy.grid.size());
	for (int i = 0; i < copy.grid.size(); i++) {
		grid.push_back(copy.grid[i]);
	}
}

SudokuGrid& SudokuGrid::transpose() {
	SudokuGrid temp = *this;
	for (int row = 0; row < pow(n,2); ++row) {
		for (int col = 0; col < pow(n,2); ++col) {
			this->grid[row][col] = temp.grid[col][row];
		}
	}
	return *this;
}

SudokuGrid& SudokuGrid::swapRows(int row1, int row2) {
	for (int col = 0; col < pow(n,2); ++col) {
		byte temp = this->grid[row1][col];
		this->grid[row1][col] = this->grid[row2][col];
		this->grid[row2][col] = temp;
	}
	return *this;
}

SudokuGrid& SudokuGrid::swapCols(int col1, int col2) {
	for (int row = 0; row < pow(n,2); ++row) {
		byte temp = this->grid[row][col1];
		this->grid[row][col1] = this->grid[row][col2];
		this->grid[row][col2] = temp;
	}
	return *this;
}

SudokuGrid& SudokuGrid::swapBands(int band1, int band2) {
	for (int offset = 0; offset < n; ++offset) {
		swapRows(n * band1 + offset, n * band2 + offset);
	}
	return *this;
}

SudokuGrid& SudokuGrid::swapStacks(int stack1, int stack2) {
	for (int offset = 0; offset < n; ++offset) {
		swapCols(n * stack1 + offset, n * stack2 + offset);
	}
	return *this;
}

SudokuGrid& SudokuGrid::permuteCols(int stackindex, int index, bool reversed, int columnindex /* = 0*/, std::stack<std::tuple<int, int> > stack) {
	//cout << "chanign cols:" << columnindex + stackindex * n << " and " <<  columnindex + index/(factorial(n - columnindex - 1)) + stackindex * n << endl;
	int swap1 = columnindex + stackindex * n;
	int swap2 = columnindex + index/(factorial(n - columnindex - 1)) + stackindex * n;
	if (reversed) {
		stack.push(tuple<int, int>(swap1, swap2));
	}
	else {
		swapCols(swap1, swap2);
	}
	if (columnindex == n - 2) {
		if (reversed) {
			while(!stack.empty()) {
				tuple<int, int> tuple = stack.top();
				stack.pop();
				swapCols(get<0>(tuple), get<1>(tuple));
			}
		}
		return *this;
	}

	else {
		int newindex = index%(factorial(n - columnindex - 1));
		return permuteCols(stackindex, newindex, reversed, ++columnindex, stack);
	}
}

SudokuGrid& SudokuGrid::permuteRows(int bandindex, int index, bool reversed, int rowindex/* = 0*/, std::stack<std::tuple<int, int> > stack) {
	//cout << "chanign rows:" << rowindex + bandindex * n << " and " <<  rowindex + index/(factorial(n - rowindex - 1)) + bandindex * n << endl;
	int swap1 = rowindex + bandindex * n;
	int swap2 = rowindex + index/(factorial(n - rowindex - 1)) + bandindex * n;
	if (reversed) {
		stack.push(tuple<int, int>(swap1, swap2));
	}
	else {
		swapRows(swap1, swap2);
	}
	if (rowindex == n - 2) {
		if (reversed) {
			while(!stack.empty()) {
				tuple<int, int> tuple = stack.top();
				stack.pop();
				swapRows(get<0>(tuple), get<1>(tuple));
			}
		}
		return *this;
	}
	else {
		int newindex = index%(factorial(n - rowindex - 1));
		return permuteRows(bandindex, newindex, reversed, ++rowindex, stack);
	}
}

SudokuGrid& SudokuGrid::permuteStacks(int index, bool reversed, int stackindex/* = 0*/, std::stack<std::tuple<int, int> > stack) {
	//cout << "chanign stacks:" << stackindex + stackindex * n << " and " <<  stackindex + index/(factorial(n - stackindex - 1)) + stackindex * n << endl;
	int swap1 = stackindex;
	int swap2 = stackindex + index/(factorial(n - stackindex - 1));
	if (reversed) {
		stack.push(tuple<int, int>(swap1, swap2));
	}
	else {
		swapStacks(swap1, swap2);
	}
	if (stackindex == n - 2) {
		if (reversed) {
			while(!stack.empty()) {
				tuple<int, int> tuple = stack.top();
				stack.pop();
				swapStacks(get<0>(tuple), get<1>(tuple));
			}
		}
		return *this;
	}

	else {
		int newindex = index%(factorial(n - stackindex - 1));
		return permuteStacks(newindex, reversed, ++stackindex, stack);
	}
}

SudokuGrid& SudokuGrid::permuteBands(int index, bool reversed, int bandindex/* = 0*/, std::stack<std::tuple<int, int> > stack) {
	//cout << "chanign bands:" << bandindex << " and " <<  bandindex + index/(factorial(n - bandindex - 1)) << endl;
	int swap1 = bandindex;
	int swap2 = bandindex + index/(factorial(n - bandindex - 1));
	if (reversed) {
		stack.push(tuple<int, int>(swap1, swap2));
	}
	else {
		swapBands(swap1, swap2);
	}
	if (bandindex == n - 2) {
		if (reversed) {
			while(!stack.empty()) {
				tuple<int, int> tuple = stack.top();
				stack.pop();
				swapBands(get<0>(tuple), get<1>(tuple));
			}
		}
		return *this;
	}

	else {
		int newindex = index%(factorial(n - bandindex - 1));
		return permuteBands(newindex, reversed, ++bandindex, stack);
	}
}

SudokuGrid& SudokuGrid::permuteBandsNonRec(int index, bool reversed, int bandindex/* = 0*/, std::stack<std::tuple<int, int> > stack) {
	//cout << "changing bands:" << bandindex << " and " <<  bandindex + index/(factorial(n - bandindex - 1)) << endl;
	int perms = factorial(n);
	int save = 0;
	for (int length = n; length > 1; --length) {
		perms /= length;
		swapBands(save, index / perms);
		save = index / perms;
		index %= perms;
	}
	return *this;
}

vector<byte> SudokuGrid::getNotZeroCells() {
	vector<byte> vec = vector<byte>();
	for (int row = 0; row < pow(n,2); ++row) {
		for (int col = 0; col < pow(n,2); ++col) {
			if (grid[row][col] != 0) {
				vec.push_back((row+1) * 10 + (col+1));
			}
		}
	}
	return vec;
}

bool SudokuGrid::operator ==(const SudokuGrid &b) const {
	if (b.n == n) {
		for (int row = 0; row < pow(n,2); ++row) {
			for (int col = 0; col < pow(n,2); ++col) {
				if ((*this).grid[row][col] != b.grid[row][col]) {
					return false;
				}
			}
		}
		return true;
	}
	return false;
}
